##################################################################################
#                                                                                #
#                       RETO MINSAIT REAL ESTATE MODELLING                       #
#                                                                                #
#                              Autores: Data Fever                               #
#                                                                                #
#                            AFI Escuela de Finanzas                             #
##################################################################################


# ------------------------ Carga de paquetes ------------------------ 

import pandas as pd
import matplotlib.pyplot as plt
import numpy as np
from PIL import Image
from os import listdir
from os.path import isfile, join
import os
from sklearn.model_selection import train_test_split


# Función de carga y parseo de datos

def load_data(file):
	# Lectura del archivo y conversión de tipos
	df = pd.read_csv(file, sep = '|', encoding = 'utf-8')
	df.HY_cod_postal = df.HY_cod_postal.astype(str)
	df.HY_id = df.HY_id.astype(str)


	##############################################################
	#           Completamos la columna de provincias             #
	##############################################################


	# Función auxiliar para encontrar la ciudad dado el código postal
	def get_city(df, cod_pos):
	    aux = df[(df.HY_cod_postal == cod_pos) & (~df.HY_provincia.isna())]
	    return pd.unique(aux.HY_provincia)

	# Función para completar la variable de provincia según el código postal  
	def fill_city(df):
	    aux = df[df.HY_provincia.isna()]
	    lost_cod_postal = pd.unique(aux.HY_cod_postal)
	    
	    for cod in lost_cod_postal:
	        city = get_city(df, cod)
	        if len(city)> 0:        
	            df.loc[(df.HY_cod_postal == cod ) & (df.HY_provincia.isna()), 'HY_provincia'] = city[0]

	    # Aún quedan dos registros: los arreglamos manualmente    
	    df.loc[df.HY_cod_postal == '03150' , 'HY_provincia'] = 'Alicante'
	    df.loc[df.HY_cod_postal == '06460' , 'HY_provincia'] = 'Badajoz'

	    return df

	df = fill_city(df)    
	    

	##############################################################
	#              Completamos las variables IDEA                #
	##############################################################

	# Función auxiliar para conseguir la media de 
	def get_IDEA_info(df, cp):
	    df = df[df.HY_cod_postal == cp]
	    return df.mean() # df.median()
	    

	def fill_IDEA(df):
	    
	    idea_cols = ['IDEA_area', 'IDEA_poblacion', 'IDEA_densidad', 'IDEA_pc_1960', 'IDEA_pc_1960_69', 
	               'IDEA_pc_1970_79', 'IDEA_pc_1980_89', 'IDEA_pc_1990_99', 'IDEA_pc_2000_10', 'IDEA_pc_comercio',
	               'IDEA_pc_industria', 'IDEA_pc_oficina', 'IDEA_pc_otros', 'IDEA_pc_residencial', 'IDEA_pc_trast_parking',
	               'IDEA_ind_tienda', 'IDEA_ind_turismo', 'IDEA_ind_alimentacion', 'IDEA_ind_riqueza', 'IDEA_rent_alquiler',
	               'IDEA_ind_elasticidad', 'IDEA_ind_liquidez', 'IDEA_unitprice_sale_residential', 'IDEA_price_sale_residential', 
	               'IDEA_stock_sale_residential', 'IDEA_demand_sale_residential', 'IDEA_unitprice_rent_residential',
	               'IDEA_price_rent_residential', 'IDEA_stock_rent_residential', 'IDEA_demand_rent_residential']

	    
	    # Ahora queremos quedarnos con los CP que aparezcan varias veces
	    aux2 = df[df.HY_cod_postal.duplicated(keep = False)]
	    
	    # Nos quedamos los códigos postales de los registros que tienen muchos NAs
	    cp_nas = aux2.loc[aux2.isna().sum(axis = 1) >= 15, 'HY_cod_postal']
	       
	    # Imputamos los NA con la función auxiliar anterior
	    for cp in cp_nas:
	        df.loc[(df.HY_cod_postal == cp) & (df[idea_cols].isna().sum(axis = 1) == len(idea_cols)), idea_cols] = get_IDEA_info(aux2, cp)[idea_cols].values

	    return df

	df = fill_IDEA(df)

	# Completamos todas las filas con su media
	numerics = ['int16', 'int32', 'int64', 'float16', 'float32', 'float64']

	for col in list(df.select_dtypes(include=numerics)):
	    df[col] = df[col].fillna(np.mean(df[col]))


	##############################################################
	#                   OBTENCIÓN DE IMÁGENES                    #
	##############################################################

	# Función para obtener el ID del anuncio a partir del path de la foto
	def get_ID(files):
	    files = [x[0:7] for x in files]
	    return files


	# Función para obtener el peso, anchura y altura de una foto
	def get_size(file):
	    try:
	        im = Image.open(file)
	        width, height = im.size
	    except:
	        width, height = [0, 0]
	    return [os.path.getsize(file),width, height]


	# Función para añadir información obtenida de las fotos. Incluyen:
	#    * NUM_FOTOS: número de fotos asociadas al anuncio
	#    * NUM_PESO_MEDIO: peso medio de las imágenes 
	#    * NUM_ALTURA_MEDIA: altura media de las imágenes
	#    * NUM_ANCHURA_MEDIA: anchura media de las imágenes
	#    * area_foto: producto de la altura y la anchura de las imágenes
	#    * calidad_foto: peso en función del área de las imágenes    
	def get_pic_info(df, path):   
	    onlyfiles = [f for f in listdir(path) if isfile(join(path, f))]
	    
	    files = pd.DataFrame(get_ID(onlyfiles), columns = ['ID'])
	    sizes = pd.DataFrame([get_size(path + '/' + file) for file in onlyfiles])
	    sizes = pd.concat([files,sizes], axis = 1)
	        
	    files = files.reset_index().groupby('ID').count().reset_index()
	    files.columns = ['ID', 'NUM_FOTOS']
	    
	    sizes = sizes.groupby('ID').mean().reset_index()
	    sizes.columns = ['ID', 'NUM_PESO_MEDIO','NUM_ANCHURA_MEDIA', 'NUM_ALTURA_MEDIA']
	    
	    files = files.merge(sizes)    
	    
	    df = df.merge(files, left_on = 'HY_id', right_on = 'ID', how = 'left')
	    del df['ID']
	    df['NUM_FOTOS'] = df['NUM_FOTOS'].fillna(0)
	    df['NUM_PESO_MEDIO'] = df['NUM_PESO_MEDIO'].fillna(0)
	    df['NUM_ALTURA_MEDIA'] = df['NUM_ALTURA_MEDIA'].fillna(0)
	    df['NUM_ANCHURA_MEDIA'] = df['NUM_ANCHURA_MEDIA'].fillna(0)
	    df['area_foto'] = df.NUM_ANCHURA_MEDIA*df.NUM_ALTURA_MEDIA
	    df['calidad_foto'] = df.NUM_PESO_MEDIO/(df.area_foto+1)
	    
	    return df


	df = get_pic_info(df, './imagenes_inmuebles_haya')


	##############################################################
	#              CONSTRUCCIÓN DE NUEVAS VARIABLES              #
	##############################################################


	# Contamos el número de caracteres de la descripción y distribución

	df.loc[df.HY_descripcion.isna(),'HY_descripcion'] = ''
	df['len_description'] = df.HY_descripcion.apply(lambda x: len(x))

	df.loc[df.HY_distribucion.isna(),'HY_distribucion'] = ''
	df['len_distribution'] = df.HY_distribucion.apply(lambda x: len(x))

	# La variable antiguedad representará el tiempo que pasó desde la construcción
	df.HY_antiguedad = 2018 - df.HY_antiguedad
	df.loc[df.HY_antiguedad == -70, 'HY_antiguedad'] = int(np.median(df.HY_antiguedad)+1)


	# Diferencia entre los metros útiles y los totales
	df['dif_metros'] = df.HY_metros_totales - df.HY_metros_utiles

	# Ratio entre el precio actual y el anterior
	df['ahorro_precio'] = df.HY_precio/(df.HY_precio_anterior+1)

	#Precio en función de los metros cuadrados útiles
	df['precio_metro_2'] = df.HY_precio/(df.HY_metros_utiles+1)

	# Número de quincenas que el anuncio estuvo en la página
	df['num_quincenas'] = df.GA_quincena_ult - df.GA_quincena_ini

	# Número de quincenas que el anuncio estuvo en la páquina penalizado si la última quincena fue la 47
	df['dif'] = (df.GA_quincena_ult - df.GA_quincena_ini)*(1+0.2*(df.GA_quincena_ult==47))

	# Quincena media en la que el anuncio estuvo activo
	df['quincena_media'] = (df.GA_quincena_ult + df.GA_quincena_ini)/2

	# Quincena del año media en la que el anuncio estuvo activo
	df['quincena_media_anual'] = ((df.GA_quincena_ult + df.GA_quincena_ini)/2 - 1) % 26 + 1

	# Ponderación del número de fotos según la interacción de la página
	df['fotos_interact'] = df.NUM_FOTOS*(100-df.GA_mean_bounce)



	# Variable para indicar si tiene certificado energético
	# Puntuamos cada una de las categorías de las eficiencias
	df['tiene_cert'] = 1
	df.loc[df.HY_cert_energ.isna(),'tiene_cert'] = 0
	df['cert_val'] = 0
	df.loc[df.HY_cert_energ == 'A','cert_val'] = 6
	df.loc[df.HY_cert_energ == 'B','cert_val'] = 5
	df.loc[df.HY_cert_energ == 'C','cert_val'] = 4
	df.loc[df.HY_cert_energ == 'D','cert_val'] = 3
	df.loc[df.HY_cert_energ == 'E','cert_val'] = 2
	df.loc[df.HY_cert_energ == 'F','cert_val'] = 1



	# Tomamos logaritmos y exponenciales de las variables más interesantes
	df['exp_q_ult'] = np.exp(df.GA_quincena_ult/47)
	df['log_distribution'] = np.exp(df.len_distribution/100)
	df['log_description'] = np.exp(df.len_description/100)
	df['log_views'] = np.log(df.GA_page_views+1)
	df['log_pc_otros'] = np.log(df.IDEA_pc_otros+1)
	df['exp_q_media'] = np.exp(df.quincena_media/47)
	df['log_bounce'] = np.log(df.GA_mean_bounce+1)
	df['log_exit'] = np.log(df.GA_exit_rate+1)
	df['log_num_q'] = np.log(df.num_quincenas+1)
	df['inv_views'] = 1/(df.GA_page_views + 1 )
	df['sqrt_views'] = np.sqrt(df.GA_page_views)

	# Ponderamos variables de medias según las visualizaciones del anuncio
	df['views_fotos'] = df.log_views * df.NUM_FOTOS
	df['views_distribution'] = df.log_views * df.len_distribution
	df['views_description'] = df.log_views * df.len_description
	df['views_fotos_interact'] = df.log_views * df.NUM_FOTOS /( df.GA_mean_bounce+1)
	df['views_fotos_no_interact'] = df.log_views * df.NUM_FOTOS / (101 -  df.GA_mean_bounce)


	# Los precios indicados como cero pasan a tomar el valor del precio anterior.
	df.loc[df.HY_precio == 0, 'HY_precio'] = df.loc[df.HY_precio == 0, 'HY_precio_anterior']


	return df


# Obtención de los datos de train y test procesados mediante la función anterior
df = load_data('./data/Modelar_UH2019.txt')	
df.to_csv('./train.csv', sep = '|', index = False, encoding = 'utf-8')

test = load_data('./data/Estimar_UH2019.txt')	
test.to_csv('./test.csv', sep = '|', index = False, encoding = 'utf-8')

# Creamos un conjunto de train y validacion para validar los modelos
X_train, X_test, y_train, y_test = train_test_split(df.loc[:,df.columns != 'TARGET'], df.TARGET, test_size=0.2, random_state=42)
X_train['TARGET'] = y_train
X_test['TARGET'] = y_test

X_train.to_csv('./train_08.csv', sep = '|', index = False, encoding = 'utf-8')
X_test.to_csv('./validacion.csv', sep = '|', index = False, encoding = 'utf-8')
