##################################################################################
#                                                                                #
#                       RETO MINSAIT REAL ESTATE MODELLING                       #
#                                                                                #
#                              Autores: Data Fever                               #
#                                                                                #
#                            AFI Escuela de Finanzas                             #
##################################################################################


# ------------------------ Carga de paquetes ------------------------ 

import numpy as np
import pandas as pd 
from sklearn.model_selection import train_test_split, RepeatedKFold, StratifiedKFold
from sklearn.model_selection import GridSearchCV, RandomizedSearchCV
from sklearn import feature_selection
from sklearn import linear_model
from sklearn import preprocessing
from sklearn import metrics
from sklearn import ensemble
from sklearn import decomposition
from sklearn import discriminant_analysis
from xgboost import XGBRegressor

# Lectura de los archivos de train y test
comptrain = pd.read_csv("train_08.csv", sep = "|")
comptest = pd.read_csv("validacion.csv", sep = "|")

# Filtrado de ouliers
train = comptrain[comptrain["TARGET"] <= 85]
test = comptest


#Selección de las columnas numéricas
Xtrain = train.drop(["TARGET", "HY_cod_postal", "HY_distribucion", "HY_descripcion", "HY_provincia", "HY_tipo", "HY_cert_energ"], axis = 1)
Ytrain = pd.DataFrame(train["TARGET"])

Xtest = test.drop(["TARGET", "HY_cod_postal", "HY_distribucion", "HY_descripcion", "HY_provincia", "HY_tipo", "HY_cert_energ"], axis = 1)
Ytest = pd.DataFrame(test["TARGET"])



# Función auxiliar para entrenar un modelo, dando predicciones y errores sobre el conjunto de test
def modelator_pred(modelo, dtr, dts, name, corte, kpis_pred = None, validation = True): 
    
    # Eliminamos outliers
    train = dtr[dtr["TARGET"] <= corte]
    test = dts 
    
    # Elminamos las columnas no numéricas y separamos variables de etiqueta
    Xtrain = train.drop(["TARGET", "HY_cod_postal", "HY_distribucion", "HY_descripcion", "HY_provincia", "HY_tipo", "HY_cert_energ"], axis = 1)
    Ytrain = pd.DataFrame(train["TARGET"])      
    
    if validation:
        Xtest = test.drop(["TARGET", "HY_cod_postal", "HY_distribucion", "HY_descripcion", "HY_provincia", "HY_tipo", "HY_cert_energ"], axis = 1)
        Ytest = pd.DataFrame(test["TARGET"])
    else:
        Xtest = test.drop(["HY_cod_postal", "HY_distribucion", "HY_descripcion", "HY_provincia", "HY_tipo", "HY_cert_energ"], axis = 1)
    
    # Entrenamiento y predicción
    modelo.fit(Xtrain, np.ravel(Ytrain))
    pred = modelo.predict(Xtest)

    if validation:
        MEDIANAE = metrics.median_absolute_error(Ytest, pred)

        # Añadimos el modelo actual
        r = kpis_pred.shape[0]
        kpis_pred.loc[r] = [name,corte,MEDIANAE]

    return pred



########### Entrenamiento de Random Forest:
#Dataframe para almacenar los resultados:
kpis_pred = pd.DataFrame({'Modelo': [],
                           'Cortes':[],
                           'MEDIANAE':[]})      

# Definición de hiperparamtros:
n = 50
param_rf = {"max_depth": [3, 5, 10, 15, 20, 25, 30, None],
            "max_features": ["sqrt", "log2", None],
            "min_samples_split": [2, 4, 8, 16, 32, 64],
            "min_samples_leaf": [1, 2, 4, 8, 16, 32, 64],
            "bootstrap": [True, False]}
          

# Definición del modelo y búsqueda de los hiperparámetros
forest = ensemble.RandomForestRegressor(n_estimators = 100, criterion='mae')
random_search = RandomizedSearchCV(forest, param_distributions = param_rf, n_iter = n, cv = 5, scoring = 'neg_median_absolute_error')
mod_rf = random_search.fit(Xtrain, np.ravel(Ytrain))
model_rf = mod_rf.best_estimator_


# Busqueda del mejor corte en los datos para definir outliers
# Se utiliza el regresor obtenido en el search anterior
for i in range(1,10):
    for n in range(70, 100):
        modelator_pred(model_rf, comptrain, comptest, "RF", n, kpis_pred)

# Agrupamos por los distrintos umbrales considerados para elegir
# el mejor punto de corte a partir el cual los datos se consideran
# outliers
medcutrf = kpis_pred.groupby("Cortes").mean()

# Entrenamos con el mejor modelo obtenido y el mejor corte



##########Gradient Boosting:
#Dataframe para almacenar los resultados:
kpis_pred = pd.DataFrame({'Modelo': [],
                           'Cortes':[],
                           'MEDIANAE':[]})     
# Definición de hiperparamtros:
n = 99
param_GB = {"learning_rate": np.logspace(-5,0,12),
            "max_depth": [3, 5, 10, 15, 20, 25, 30, None],
            "min_samples_split": [2, 4, 8, 16, 32, 64],
            "min_samples_leaf": [1, 2, 4, 8, 16, 32, 64],
            "subsample" : [0.1, 0.2, 0.5, 0.9, 1],
            "max_features": ["sqrt", "log2", None]}
            

# Definición del modelo y búsqueda de los hiperparámetros
gb = ensemble.GradientBoostingRegressor(loss = "quantile", alpha=0.5)
random_search = RandomizedSearchCV(gb, param_distributions=param_GB, n_iter=n, cv=5, scoring = 'neg_median_absolute_error')
mod_gb = random_search.fit(Xtrain, np.ravel(Ytrain))
model_gb = mod_gb.best_estimator_

# Busqueda del mejor corte en los datos para definir outliers
# Se utiliza el regresor obtenido en el search anterior
for i in range(1,10):
    for n in range(70,100):
        modelator_pred(model_gb, comptrain, comptest, "GB", n, kpis_pred)

# Determinamos el mejor umbral        
medcutgb = kpis_pred.groupby("Cortes").mean()



# Entrenamos sobre el conjunto total de train y predecimos test
train = pd.read_csv("train.csv", sep = "|")
test = pd.read_csv("test.csv", sep = "|")

# Obtención etiquetas de RandomForest
predictionsrf = modelator_pred(model_rf, train, test, "RF", 76, validation = False)
predictionsrf = predictionsrf.tolist()

# Obtención etiquetas de GXBoost
predictionsgb = modelator_pred(model_gb, train, test, "GB", 96, validation = False)
predictionsgb = predictionsgb.tolist()


# Guardamos las predicciones en un archivo CSV
predictions = pd.DataFrame(data={"GB": predictionsgb, "RF": predictionsrf})
predictions.to_csv("./predictions.csv", sep=',',index=True)